/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Angle.hpp
 *
 *  Created on: 2010/11/18
 *      Author: HARK Development Team
 */

#ifndef DIMPRO_ANGLE_HPP_
#define DIMPRO_ANGLE_HPP_

#include <cmath>
#include "scalar.hpp"

namespace NSDimensionPromise
{

/**
 * Angleの次元定義.
 */
template <typename T>
struct AngleBaseTraits
{
    typedef ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0> ScalarType;
};

/**
 * Angleを表すベースクラス.
 */
template <typename T>
class AngleBase : public AngleBaseTraits<T>::ScalarType
{
public:
    /**
     * Angleを表すスカラー変数を生成する.
     * @param radian 変数の初期値
     */
    explicit AngleBase(const T& radian = T())
        : AngleBaseTraits<T>::ScalarType(radian)
    {  }

    /**
     * Angleを表すスカラー変数をコピーして生成する.
     * @param rhs コピー元の変数
     */
    AngleBase(const typename AngleBaseTraits<T>::ScalarType& rhs)
        : AngleBaseTraits<T>::ScalarType(rhs)
    {  }

    /**
      * ラジアン単位の値を得る.
      * @return ラジアン
      */
    const T& ToRadian() const
    {
        return this->X();
    }

    /**
      * 度数単位の値を得る.
      * @return 度数
      */
    const T ToDegree() const
    {
        return static_cast<T>(this->X() * 180.0 / M_PI);
    }

    /**
      * ラジアンからAngleを表すスカラー変数を生成する.
      * @return Angleを表すスカラー変数
      */
    static AngleBase FromRadian(const T& radian)
    {
        return AngleBase(radian);
    }

    /**
      * 度数からAngleを表すスカラー変数を生成する.
      * @return Angleを表すスカラー変数
      */
    static AngleBase FromDegree(const T& degree)
    {
        return AngleBase(degree * M_PI / 180.0);
    }
};

/**
 * よく使うAngleの型.
 */
typedef AngleBase<NormalValueType> Angle;


using std::sin;

template <typename T>
inline const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
sin(const AngleBase<T>& angle)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(sin(angle.ToRadian()));
}

using std::asin;

template <typename T>
inline const AngleBase<T>
asin(const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& ratio)
{
    return AngleBase<T>(asin(ratio.X()));
}

using std::cos;

template <typename T>
inline const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
cos(const AngleBase<T>& angle)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(cos(angle.ToRadian()));
}

using std::acos;

template <typename T>
inline const AngleBase<T>
acos(const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& ratio)
{
    return AngleBase<T>(acos(ratio.X()));
}

using std::tan;

template <typename T>
inline const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
tan(const AngleBase<T>& angle)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(tan(angle.ToRadian()));
}

using std::atan;

template <typename T>
inline const AngleBase<T>
atan(const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& ratio)
{
    return AngleBase<T>(atan(ratio.X()));
}

}

#endif /* DIMPRO_ANGLE_HPP_ */
