/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * DirectionPassFilter.hpp
 *
 *  Created on: 2010/11/15
 *      Author: HARK Development Team
 */

#ifndef DIRECTIONPASSFILTER_HPP_
#define DIRECTIONPASSFILTER_HPP_

#include <cstddef>

#include <DimensionPromise/types.hpp>
#include <SeparationSetting.hpp>
#include <MaxMin.hpp>
#include <InterauralDifference.hpp>
#include <Channel.hpp>
#include <Signal.hpp>
#include <LRSignal.hpp>

namespace NSHarkBinaural
{

/**
 * theta±widthの上限値・下限値を計算する.
 * @param theta 中心の角度
 * @param width 角度の範囲
 * @param mma 出力用。計算した上限値・下限値（-90°以上、90°以下の角度）を格納する。
 */
void CalculateMaxMin(
    NSDimensionPromise::Angle theta,
    NSDimensionPromise::Angle width,
    MaxMin<NSDimensionPromise::Angle>& mma/*, MaxMin<int>& mmi*/);

/**
 * 周波数遮断・通過フラグ設定モード.
 */
struct PassBandSetMode
{
    typedef enum {
        /**
         * 遮断情報は設定するが、通過情報は設定しない.
         * このモードでフラグを設定すると、必ず遮断周波数帯の数が増える。
         */
        CutOffOnly,

        /**
         * 通過情報は設定するが、遮断情報は設定しない.
         * このモードでフラグを設定すると、必ず通過周波数帯の数が増える。
         */
        PassOnly,

        /**
         * 通過情報も、遮断情報も設定する.
         */
        CutOffAndPass
    } Kind;
};

/**
 * 周波数の遮断・通過を表すフラグ.
 */
class PassBandFlag
{
public:
    typedef enum {
        /**
         * 遮断.
         */
        CutOff,

        /**
         * 通過.
         */
        Pass
    } Kind;

    /**
     * 遮断・通過フラグを生成する.
     * @param flag フラグの初期値
     */
    PassBandFlag(Kind flag = Pass)
        : flag(flag)
    { }

    /**
     * フラグを設定する.
     * @param setMode フラグ設定モード
     * @param flag 設定したい遮断・通過フラグの値
     */
    void Flag(PassBandSetMode::Kind setMode, Kind flag);

    /**
     * フラグを取得する.
     * @return フラグの値
     */
    Kind Flag()
    {
        return this->flag;
    }

private:
    /**
     * 通過・遮断を表すフラグ.
     */
    Kind flag;
};

struct IIPDS;

/**
 * 周波数通過帯域.
 * どの周波数を遮断・通過させるかを表すクラス
 */
class PassBand
{
public:
    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<PassBand> ptr_t;

    /**
     * 周波数通過帯域を生成する.
     * @param size 周波数の種類数
     */
    PassBand(size_t size)
        : passBand(size)
    { }

    /**
     * エピポーラの方法により、通過帯域を設定.
     * @param setMode 通過・遮断情報設定モード
     * @param separationSetting 分離の設定
     * @param dir 分離方向
     * @param range 分離範囲
     * @param ipd 両耳間位相差配列
     * @param signalPowerSpectrum 音声信号のパワースペクトル
     * @param roomnoisePowerSpectrum ルームノイズのパワースペクトル
     * @param sourceDistance 頭部の中心から信号源までの距離
     * @param headRadius 頭部の半径
     */
    void SetPassBandEpipolar(
        PassBandSetMode::Kind setMode,
        SeparationSetting::ptr_t separationSetting,
        NSDimensionPromise::Angle dir, NSDimensionPromise::Angle range,
        IPD::ptr_t ipd,
        Unspectrum::ptr_t signalPowerSpectrum, Unspectrum::ptr_t roomnoisePowerSpectrum,
        NSDimensionPromise::Length sourceDistance, NSDimensionPromise::Length headRadius);

    /**
     * HRTFを用いて、通過帯域を設定.
     * @param setMode 通過・遮断情報設定モード
     * @param separationSetting 分離の設定
     * @param dir 分離方向
     * @param range 分離範囲
     * @param ipd 両耳間位相差配列
     * @param iid 両耳間強度差配列
     * @param signalPowerSpectra 音声信号のパワースペクトル（デシベル値）
     * @param roomnoisePowerSpectra ルームノイズのパワースペクトル（デシベル値）
     * @param iipds 頭部伝達関数
     * @param thresDB ルームノイズに対する信号パワーの閾値
     */
    void SetPassBandHRTF(
        PassBandSetMode::Kind setMode,
        SeparationSetting::ptr_t separationSetting,
        NSDimensionPromise::Angle dir, NSDimensionPromise::Angle range,
        IPD::ptr_t ipd, IID::ptr_t iid,
        LRUnspectra::ptr_t signalPowerSpectra, LRUnspectra::ptr_t roomnoisePowerSpectra,
        IIPDS& iipds,
        Unspectrum::ElementType thresDB);

    /**
     * 指定された周波数帯の通過・遮断情報への参照を得る.
     * @param index 周波数帯を表すインデックス
     * @return 通過・遮断情報
     */
    PassBandFlag& operator [](size_t index)
    {
        return this->passBand[index];
    }

    /**
     * 指定された周波数帯の通過・遮断情報への参照を得る.
     * @param index 周波数帯を表すインデックス
     * @return 通過・遮断情報
     */
    const PassBandFlag& operator [](size_t index) const
    {
        return this->passBand[index];
    }

    /**
     * 指定された周波数帯の通過・遮断情報への参照を得る.
     * @param index 周波数帯を表すインデックス
     * @return 通過・遮断情報
     */
    PassBandFlag& At(size_t index)
    {
        return this->passBand.at(index);
    }

    /**
     * 指定された周波数帯の通過・遮断情報への参照を得る.
     * @param index 周波数帯を表すインデックス
     * @return 通過・遮断情報
     */
    const PassBandFlag& At(size_t index) const
    {
        return this->passBand.at(index);
    }

    /**
     * 通過・遮断情報を保持している周波数帯の数を得る.
     * @return 周波数帯の数
     */
    size_t Size() const
    {
        return this->passBand.size();
    }

private:
    /**
     * 周波数領域における、通過・遮断フラグの配列.
     * 実際の1要素に対応する周波数の幅は、信号のサンプリング周波数と
     * 離散フーリエ変換のサンプル数に依存する
     */
    std::vector<PassBandFlag> passBand;
};

/**
 * 遮断・通過フラグ配列の全要素に値を設定する.
 * フラグ配列の初期化に用いる
 * @param passBand 遮断・通過フラグ配列
 * @param flag 設定するフラグの値
 */
void SetPassBandFlags(PassBand::ptr_t passBand, PassBandFlag::Kind flag);

/**
 * 周波数帯抽出機.
 * スペクトルから通過帯だけ抜き出す機能を提供するクラス
 */
class BandExtractor
{
public:
    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<BandExtractor> ptr_t;

    /**
     * 左右スペクトル型
     */
    typedef ChannelHolder<Spectrum::ptr_t, Spectrum::ptr_t> LRSpectra;

    /**
     * 周波数帯抽出機を生成する.
     * @param lrspectra データ抜き出し対象の2チャンネルのスペクトル
     */
    BandExtractor(LRSpectra::ptr_t lrspectra)
        : lrspectra(lrspectra)
    { }

    /**
     * 通過帯情報にしたがってスペクトルデータを抜き出す.
     * @param passBand 通過帯情報
     * @return 抜き出したスペクトル 入力スペクトルと通過帯情報の長さが異なるならnull
     */
    LRSpectra::ptr_t Extract(PassBand::ptr_t passBand);
private:

    /**
     * 抜き出し対象のスペクトル
     */
    LRSpectra::ptr_t lrspectra;
};

}

#endif /* DIRECTIONPASSFILTER_HPP_ */
